import inspect
import networkx as nx
from mmodel.draw import draw_graph
from copy import deepcopy


class ModelGraph(nx.DiGraph):

    """Create model graphs

    ModelGraph inherits from `networkx.DiGraph()`, which has all `DiGraph`
    methods.

    The class adds the "type" attribute to the graph attribute. The factory method
    returns a copy of the dictionary. It is equivalent to
    ``{"type": "ModelGraph"}.copy()`` when called.

    The additional graph operations are added:
    - add_grouped_edges and set_node_objects.
    - Method add_grouped_edges, cannot have both edges to be a list.
    - Method set_node_object update nodes with the node callable "func" and returns.
    - The method adds callable signature 'sig' to the node attribute
    """

    graph_attr_dict_factory = {"type": "ModelGraph"}.copy

    def set_node_object(self, node, func, returns, modifiers: list = []):
        """Add or update the functions of existing node

        If the node does not exist, create the node.
        In the end, the edge attributes are re-determined

        Modifiers are applied directly onto the node.
        """

        # if node not in self.nodes:
        #     self.add_node(node)

        node_dict = self.nodes[node]

        # store the base object
        node_dict["base_obj"] = func

        for mod in modifiers:
            func = mod(func)

        sig = inspect.signature(func)
        node_dict.update(
            {"func": func, "sig": sig, "returns": returns, "modifiers": modifiers}
        )
        self.update_graph()

    def set_node_objects_from(self, node_objects: list):
        """Update the functions of existing nodes

        The method is the same as add node object
        """

        for node_obj in node_objects:
            # unzipping works for input with or without modifiers
            self.set_node_object(*node_obj)

    def view_node(self, node: str):
        """view node information

        The node information is kept consistent with the graph information.
        """

        node_dict = self.nodes[node]
        sig_list = [str(param) for param in node_dict["sig"].parameters.values()]
        # if it is not a proper function just print the repr
        # Model class instance has the attr __name__
        base_name = getattr(node_dict["base_obj"], "__name__", repr(callable))
        mod_str_list = [
            getattr(mod, "info", mod.__name__) for mod in node_dict["modifiers"]
        ]

        if mod_str_list:
            mod_str = ", ".join(mod_str_list)
        else:
            mod_str = "none"

        return "\n".join(
            [
                f"{node} node",
                f"  base callable: {base_name}",
                f"  signature: {', '.join(sig_list)}",
                f"  returns: {', '.join(node_dict['returns'])}",
                f"  modifiers: {mod_str}",
            ]
        )

    def add_grouped_edge(self, u, v):
        """Add linked edge

        For mmodel, a group edge (u, v) allows u or v
        to be a list of nodes. Represents several nodes
        flow into one node or the other way around.
        """

        if isinstance(u, list) and isinstance(v, list):
            raise Exception("only one edge node can be a list")

        # use add edges from to run less update graph
        # currently a compromise
        if isinstance(u, list):
            self.add_edges_from([(_u, v) for _u in u])
        elif isinstance(v, list):
            self.add_edges_from([(u, _v) for _v in v])
        else:  # neither is a list
            self.add_edge(u, v)

    def add_grouped_edges_from(self, group_edges: list):
        """Add edges from grouped values"""

        for u, v in group_edges:
            self.add_grouped_edge(u, v)

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """Modify add_edge to update the edge attribute in the end"""

        super().add_edge(u_of_edge, v_of_edge, **attr)
        self.update_graph()

    def add_edges_from(self, ebunch_to_add, **attr):
        """Modify add_edges_from to update the edge attributes"""

        super().add_edges_from(ebunch_to_add, **attr)
        self.update_graph()

    def update_graph(self):
        """Update edge attributes based on node objects and edges"""

        for u, v in self.edges:
            u_rts = set(self.nodes[u].get("returns", ()))
            v_sig = self.nodes[v].get("sig", None)

            if v_sig is not None:
                v_params = set(v_sig.parameters.keys())
                self.edges[u, v]["val"] = list(u_rts.intersection(v_params))

    def __str__(self):
        """Output graph information"""
        # default string is the string output of networkx.Graph
        default_str = "".join(
            [
                type(self).__name__,
                f" named {self.name!r}" if self.name else "",
                f" with {self.number_of_nodes()} nodes and {self.number_of_edges()} edges",
            ]
        )
        docstring = self.graph.get("doc", "")

        return f"{default_str.rstrip()}\n\n{docstring}".rstrip()

    def draw(self, method: callable = draw_graph):
        """Draw the graph

        A drawing is provided. Defaults to ``draw_graph``
        """

        return method(self, str(self))

    def deepcopy(self):
        """Deepcopy graph

        The graph.copy method is a shallow copy. Deepcopy creates copy for the
        attributes dictionary.
        `graph.copy<https://networkx.org/documentation/stable/reference/classes
        /generated/networkx.Graph.copy.html>_`

        However, for subgraphs, deepcopy is incredibly inefficient because
        subgraph contains '_graph', which stores the original graph.
        An alternative method is to copy the code from the copy method,
        but use deepcopy for the items.
        """
        # return deepcopy(self)

        G = self.__class__()
        G.graph.update(deepcopy(self.graph))
        G.add_nodes_from((n, deepcopy(d)) for n, d in self._node.items())
        G.add_edges_from(
            (u, v, deepcopy(datadict))
            for u, nbrs in self._adj.items()
            for v, datadict in nbrs.items()
        )

        return G



"""Filters that used to create subgraph"""
import networkx as nx


def subgraph_by_parameters(graph, parameters: list):
    """Construct subgraph based on parameters

    If a parent node is included, so are the child nodes.

    :return: subgraph view of the filtered graph
    """

    subgraph_nodes = []

    for node, sig in nx.get_node_attributes(graph, "sig").items():
        sig_params = sig.parameters
        for param in parameters:
            if param in sig_params:
                subgraph_nodes.append(node)
                subgraph_nodes.extend(nx.descendants(graph, node))

    return graph.subgraph(subgraph_nodes)


def subgraph_by_nodes(graph, nodes: list):
    """Construct subgraph based on nodes

    :return: subgraph view of the filtered graph
    """

    return graph.subgraph(nodes)


def subgraph_by_returns(graph, returns: list):
    """Construct subgraph based on node returns

    :return: subgraph view of the filtered graph

    For mmodel graphs, returns from all the internal nodes are unique.
    Therefore the function only checks if function nodes overlap with
    the target return list. If a child node is included, so are the
    parent nodes.
    """

    subgraph_nodes = []

    for node, rts in nx.get_node_attributes(graph, "returns").items():

        if not set(rts).isdisjoint(returns):  # check if they overlap
            subgraph_nodes.append(node)
            subgraph_nodes.extend(nx.ancestors(graph, node))

    return graph.subgraph(subgraph_nodes)



import inspect
import networkx as nx


def param_sorter(parameter):
    """Sorter for argument parameter

    The values in the tuple are compared in sequential order
    1. Order by parameter kind
    2. Default parameter rank at the end of its kind
    3. Alphabetical order

    :param inspect.Parameter parameter: parameter object
    :rtype: (bool, parameter.name, parameter.kind)
    """

    if parameter.default is not parameter.empty:
        return parameter.kind, True, parameter.name
    else:
        return parameter.kind, False, parameter.name


def model_signature(graph):
    """Obtain the signature from the model graph

    :param DiGraph graph: networkx.Digraph() object,
        with 'func_signature', 'returns' defined for nodes
        and "parameters" for edges.
        The args are a dictionary of inspected signature
    """

    parameters = {}
    for sig in nx.get_node_attributes(graph, "sig").values():
        for pname, param in sig.parameters.items():
            if pname in parameters:
                if param_sorter(parameters[pname]) >= param_sorter(param):
                    continue
            parameters.update({pname: param})

    for returns in nx.get_node_attributes(graph, "returns").values():
        for rt in returns:
            parameters.pop(rt, None)  # if doesn't exist return None

    return inspect.Signature(sorted(parameters.values(), key=param_sorter))


def model_returns(graph):
    """Obtain the return parameter from the model graph

    The assumption is that all return parameter names are unique.
    """

    returns = []

    for node in graph.nodes():
        if graph.out_degree(node) == 0:
            returns.extend(graph.nodes[node]["returns"])

    returns.sort()
    return returns


def replace_signature(signature, replacement_dict):
    """Replace signature with a dictionary of (key, pair)

    The function is used to replace several input parameters with an object.
    The signature is the original signature.
    The dictionary key should be the replacement object, and the values
    should be a list of the target parameters to be replaced.
    """

    params = dict(signature.parameters)
    for func, target_list in replacement_dict.items():
        for target in target_list:
            del params[target]
        params[func] = inspect.Parameter(func, 1)

    return signature.replace(parameters=sorted(params.values(), key=param_sorter))


def graph_topological_sort(graph):
    """Determine the topological order

    `nx.topological_generations` outputs a generator with each node list generation.
    However, it does not carry the node attributes. The method
    outputs a list of nodes for each generation.

    :return: topological order of the graph. Returns a list of nodes and its
        attribute
    :rtype: list

    """

    topological_order = []

    for node in nx.topological_sort(graph):
        topological_order.append((node, graph.nodes[node]))

    return topological_order


def param_counter(graph, extra_returns):
    """Count the number of times a parameter is used for graph execution

    Count all function signature parameters. For extra returns,
    add one to each count value.

    :param list extra_returns: the intermediate parameter to return for
        output.

    :return: dictionary with parameter_name: count pair
    :rtype: dict
    """

    value_list = []
    for sig in nx.get_node_attributes(graph, "sig").values():
        value_list.extend(sig.parameters.keys())

    # add the additional parameter to list
    value_list += extra_returns

    count = {}
    for value in value_list:
        count[value] = count.get(value, 0) + 1

    return count


# def loop_signature(signature, parameters):
#     """Change parameter default from scalar to list"""

#     # reset the default to a list
#     sig_param = dict(signature.parameters)
#     for parameter in parameters:
#         _param = sig_param[parameter]
#         if _param.default != inspect.Parameter.empty:
#             sig_param[parameter] = inspect.Parameter(
#                 _param.name, _param.kind, default=[_param.default]
#             )
#     return inspect.Signature(sig_param.values())


def modify_subgraph(
    graph, subgraph, subgraph_name, subgraph_obj, subgraph_returns=None
):
    """Redirect graph based on subgraph

    Find all parent nodes, not in the subgraph but child nodes in the
    subgraph. (All child nodes of subgraph nodes are in the subgraph).
    The edge attribute is passed down to the new edge. Here a new graph is
    created by deep copy the original graph

    :param graph model_graph: model_graph to modify
    :param graph subgraph: subgraph that is being replaced by a node
        subgraph is a view of the original graph
    :param str subgraph_name: name of the subgraph
    """

    graph = graph.deepcopy()

    # since we are not storing subgraph information here
    # and we are not modifying the subgraph, we do not need to copy
    # the subgraph
    # subgraph = subgraph.deepcopy()
    # subgraph.graph = {'name': subgraph_name}

    new_edges = []
    for node in subgraph.nodes():
        for parent in graph.predecessors(node):
            if parent not in subgraph:
                new_edges.append((parent, subgraph_name))
        for child in graph.successors(node):
            if child not in subgraph:
                new_edges.append((subgraph_name, child))

    graph.remove_nodes_from(subgraph.nodes)
    # remove unique edges
    graph.add_edges_from(set(new_edges))

    # subgraph requires to have the returns attribute
    if subgraph_returns is None:
        try:
            subgraph_returns = subgraph_obj.returns
        except AttributeError:
            raise Exception("'subgraph_returns' not defined")

    graph.set_node_object(subgraph_name, func=subgraph_obj, returns=subgraph_returns)
    # add the subgraph attribute to node

    return graph


def modify_node(graph, node, modifiers, node_returns=None):
    """Add modifiers to node

    The result is a new graph with node object modified.
    """

    graph = graph.deepcopy()
    func = graph.nodes[node]["func"]
    returns = node_returns or graph.nodes[node]["returns"]
    graph.set_node_object(node, func=func, returns=returns, modifiers=modifiers)

    return graph


def parse_input(signature, *args, **kwargs):
    """parse argument based on signature and input

    The default value is automatically filled.
    """

    values = signature.bind(*args, **kwargs)
    values.apply_defaults()
    return values.arguments


def is_node_attr_defined(graph, attr: str):
    """Check if all graph nodes have the target attribute defined

    Use ``set`` to ignore the order. Returns true if all nodes have the target
    attribute
    """

    return set(nx.get_node_attributes(graph, attr).keys()) == set(graph.nodes)


def is_edge_attr_defined(graph, attr: str):
    """Check if all graph edges have the target attribute defined

    Use ``set`` to ignore the order. Returns true if all nodes have the target
    attribute
    """

    return set(nx.get_edge_attributes(graph, attr).keys()) == set(graph.edges)



import graphviz
from copy import deepcopy

DEFAULT_SETTINGS = {
    "graph_attr": {
        "labelloc": "t",
        "labeljust": "l",
        "splines": "ortho",
        "ordering": "out",
    },
    "node_attr": {"shape": "box"},
}


def update_settings(label: str):
    """Update graphviz settings

    Creates a copy of the default dictionary
    and update the graph label in the graph attribute.
    """

    # copy() is shallow, does not copy the nested dict
    new_settings = deepcopy(DEFAULT_SETTINGS)
    new_settings["graph_attr"].update({"label": label})

    return new_settings


def draw_plain_graph(G, label=""):
    """Draw plain graph

    :param str name: name of the graph
    :param str label: title of the graph

    Plain graph contains the graph label (name + doc)
    Each node only shows the node name
    """

    settings = update_settings(label)
    dot_graph = graphviz.Digraph(name=G.name, **settings)

    for node in G.nodes:
        dot_graph.node(node)

    for u, v in G.edges:
        dot_graph.edge(u, v)

    return dot_graph


def draw_graph(G, label: str = ""):
    """Draw detailed graph

    :param str name: name of the graph
    :param str label: title of the graph

    Each node shows node label (name + signature + returns)
    """

    settings = update_settings(label)

    dot_graph = graphviz.Digraph(name=G.name, **settings)

    for node, ndict in G.nodes(data=True):

        if "func" in ndict:
            label = (
                f"{node}\l\n{ndict['func'].__name__}"
                f"{ndict['sig']}\lreturn {', '.join(ndict['returns'])}\l"
            )
        else:
            label = node
        dot_graph.node(node, label=label)

    for u, v, edict in G.edges(data=True):

        if "val" in edict:
            xlabel = ", ".join(edict["val"])
        else:
            xlabel = ""

        dot_graph.edge(u, v, xlabel=xlabel)

    return dot_graph



from functools import wraps
import inspect
from mmodel.utility import parse_input


def loop_modifier(parameter: str):
    """Loop - iterates one given parameter

    :param list parameter: target parameter to loop
    """

    def wrapper(func):
        @wraps(func)
        def loop_wrapped(**kwargs):

            loop_values = kwargs.pop(parameter)
            return [func(**kwargs, **{parameter: value}) for value in loop_values]

        # signature = inspect.signature(func)
        # loop_sig = loop_signature(signature, [parameter])
        # loop_wrapped.__signature__ = loop_sig

        return loop_wrapped

    wrapper.info = f"loop_modifier({parameter})"

    return wrapper


def zip_loop_modifier(parameters: list or str):
    """Pairwise loop - iterates the values from loop

    :param list or string parameters: list of the parameter to loop
        only one parameter is allowed. If string of parameters are
        provided, the parameters should be delimited by ", ".
    """

    if isinstance(parameters, str):
        parameters = parameters.split(", ")

    def wrapper(func):
        @wraps(func)
        def loop_wrapped(**kwargs):

            loop_values = [kwargs.pop(param) for param in parameters]

            result = []
            for value in zip(*loop_values):  # unzip the values

                loop_value_dict = dict(zip(parameters, value))
                rv = func(**kwargs, **loop_value_dict)
                result.append(rv)

            return result

        # signature = inspect.signature(func)
        # loop_sig = loop_signature(signature, parameters)
        # loop_wrapped.__signature__ = loop_sig

        return loop_wrapped

    wrapper.info = f"zip_loop_modifier({', '.join(parameters)})"
    return wrapper


def signature_modifier(signature_parameters):
    """Replace node object signature

    :param list signature_parameters: signature parameters to replace the original
        signature. The parameters are assumed to be "POSITIONAL_OR_KEYWORD", and no
        default values are allowed. The signature will replace the original signature
        in order.

    .. Note::

        The modifier does not work with functions that have positional only
        input parameters. When the signature_parameter length is smaller than the
        list of original signatures, there are two cases:
        1. The additional parameters have a default value - they do not show up
        in the signature, but the default values are applied to function
        2. The additional parameters do not have a default value - error is thrown
        for missing input

        ``mmodel`` allows the first case scenario, no checking is performed.

    """
    sig = inspect.Signature([inspect.Parameter(var, 1) for var in signature_parameters])

    def wrapper(func):

        old_parameters = list(inspect.signature(func).parameters.keys())

        if len(signature_parameters) > len(old_parameters):
            raise Exception(
                "The number of signature modifier parameters "
                "exceeds that of function's parameters"
            )

        # once unzipped to return iterators, the original variable returns none
        param_pair = list(zip(old_parameters, signature_parameters))

        @wraps(func)
        def wrapped(**kwargs):
            # replace keys
            replace_kwargs = {old: kwargs[new] for old, new in param_pair}
            return func(**replace_kwargs)

        wrapped.__signature__ = sig

        return wrapped

    wrapper.info = f"signature_modifier{sig}"
    return wrapper


def signature_binding_modifier():
    """Add parameter binding and checking for function

    The additional wrapper is unnecessary, but to keep a consistent
    modifier syntax. The modifier can be used on wrapped functions
    that do not have a parameter binding step (ones that only allow
    keyword arguments).

    The parse_input method binds the input args and kwargs and fills
    default values automatically. The resulting function behaves the
    same as a python function.
    """

    def wrapper(func):

        sig = inspect.signature(func)

        @wraps(func)
        def wrapped(*args, **kwargs):

            parsed_kwargs = parse_input(sig, *args, **kwargs)

            return func(**parsed_kwargs)

        return wrapped

    wrapper.info = "signature_binding_modifier"
    return wrapper



from abc import ABCMeta, abstractmethod
import h5py
from functools import partialmethod

from mmodel.utility import (
    model_signature,
    model_returns,
    graph_topological_sort,
    param_counter,
)


class TopologicalHandler(metaclass=ABCMeta):
    """Base class for layered execution following topological generation

    Data instance is used for the execution instead of attributes.
    This makes the pipeline cleaner and better for testing. A data
    instance can be a dictionary, tuple of dictionaries, or file instance.
    The Data instance is discarded after each execution or when an exception
    occurs.
    """

    def __init__(self, graph, extra_returns: list = []):

        self.__signature__ = model_signature(graph)
        self.returns = sorted(model_returns(graph) + extra_returns)
        self.order = graph_topological_sort(graph)
        self.graph = graph

    def __call__(self, **kwargs):
        """Execute graph model by layer"""

        data_instance = self.initiate(**kwargs)

        for node, node_attr in self.order:
            try:
                self.run_node(data_instance, node, node_attr)
            except Exception as e:
                self.raise_node_exception(data_instance, node, node_attr, e)

        return self.finish(data_instance, self.returns)

    @abstractmethod
    def initiate(self, **kwargs):
        """Initiate the execution"""

    @abstractmethod
    def run_node(self, data_instance, node, node_attr):
        """Run individual node"""

    @abstractmethod
    def raise_node_exception(self, data_instance, node, node_attr, e):
        """Raise exception when there is a node failure"""

    @abstractmethod
    def finish(self, data_instance, returns):
        """Finish execution"""


class MemHandler(TopologicalHandler):
    """Default model of mmodel module

    Model execute the graph by layer. For parameters, a counter is used
    to track all value usage. At the end of each layer, the parameter is
    deleted if the counter reaches 0.
    """

    def __init__(self, graph, extra_returns: list = []):
        """Add counter to the object"""
        super().__init__(graph, extra_returns)
        self.counter = param_counter(graph, extra_returns)

    def initiate(self, **kwargs):
        """Initiate the value dictionary"""

        count = self.counter.copy()
        return kwargs, count

    def run_node(self, data_instance, node, node_attr):
        """Run node

        At the end of each node calculation, the counter is updated.
        If the counter value is zero, the value is deleted.
        """
        value_dict, count = data_instance
        parameters = node_attr["sig"].parameters
        kwargs = {key: value_dict[key] for key in parameters}
        func_output = node_attr["func"](**kwargs)

        returns = node_attr["returns"]
        if len(returns) == 1:
            value_dict[returns[0]] = func_output
        else:
            value_dict.update(dict(zip(returns, func_output)))

        for key in parameters:
            count[key] -= 1
            if count[key] == 0:
                del value_dict[key]

    def raise_node_exception(self, data_instance, node, node_attr, e):
        """Raise exception

        Delete intermediate attributes
        """

        raise Exception(f"Exception occurred for node {node, node_attr}") from e

    def finish(self, data_instance, returns):
        """Finish and return values"""
        value_dict = data_instance[0]
        if len(returns) == 1:
            return_val = value_dict[returns[0]]
        else:
            return_val = tuple(value_dict[rt] for rt in returns)

        return return_val


class PlainHandler(TopologicalHandler):
    """A fast and bare-bone model

    The method stores all intermediate values in memory. The calculation steps
    are very similar to Model.
    """

    def initiate(self, **kwargs):
        """Initiate the value dictionary"""

        return kwargs

    def run_node(self, value_dict, node, node_attr):
        """Run node
        
        Store all values in a dictionary.
        """

        parameters = node_attr["sig"].parameters
        kwargs = {key: value_dict[key] for key in parameters}
        func_output = node_attr["func"](**kwargs)

        returns = node_attr["returns"]
        if len(returns) == 1:
            value_dict[returns[0]] = func_output
        else:
            value_dict.update(dict(zip(returns, func_output)))

    def raise_node_exception(self, value_dict, node, node_attr, e):
        """Raise exception

        Delete intermediate attributes
        """

        raise Exception(f"Exception occurred for node {node, node_attr}") from e

    def finish(self, value_dict, returns):
        """Finish and return values"""
        if len(returns) == 1:
            return_val = value_dict[returns[0]]
        else:
            return_val = tuple(value_dict[rt] for rt in returns)

        return return_val


class H5Handler(TopologicalHandler):
    """Model that stores all data in h5 file

    Each entry of the h5 file is unique, with the instance id, instance name
    and execution number
    """

    def __init__(self, graph, h5_filename: str, extra_returns: list = []):

        # check if file exist
        # write id attribute
        self.h5_filename = h5_filename
        self.exe_count = 0
        self.info = f"H5Handler({h5_filename})"

        super().__init__(graph, extra_returns)

    def initiate(self, **kwargs):
        """Initiate dictionary value"""
        self.exe_count += 1

        f = h5py.File(self.h5_filename, "a")

        exe_group_name = f"{id(self)}_{self.exe_count}"
        exe_group = f.create_group(exe_group_name)

        # write input dictionary to
        self.write(kwargs, exe_group)

        return f, exe_group

    def run_node(self, data_instance, node, node_attr):
        """Run node"""

        exe_group = data_instance[1]

        parameters = node_attr["sig"].parameters
        kwargs = {key: self.read(key, exe_group) for key in parameters}

        func_output = node_attr["func"](**kwargs)

        returns = node_attr["returns"]
        if len(returns) == 1:
            self.write({returns[0]: func_output}, exe_group)
        else:
            self.write(dict(zip(returns, func_output)), exe_group)

    def finish(self, data_instance, returns):
        """output parameters based on returns"""

        f, exe_group = data_instance

        if len(returns) == 1:
            rt = self.read(returns[0], exe_group)
        else:
            rt = tuple(self.read(rt, exe_group) for rt in returns)

        f.close()

        return rt

    def raise_node_exception(self, data_instance, node, node_attr, e):
        """Raise exception when exception occurred for a specific node

        The error message is written as a "_node" attribute to the current group.
        """
        f, exe_group = data_instance
        msg = f"{type(e).__name__} occurred for node {node, node_attr}: {e}"
        exe_group.attrs["note"] = msg
        f.close()

        raise Exception(f"Exception occurred for node {node, node_attr}") from e

    @staticmethod
    def write(value_dict, group):
        """Write h5 dataset/attribute by group

        :param dict value_dict: dictionary of values to write
        :param h5py.group group: open h5py group object
        """

        for k, v in value_dict.items():
            group.create_dataset(k, data=v)

    @staticmethod
    def read(key, group):
        """Read dataset/attribute by group

        :param str key: value name
        :param h5py.group group: open h5py group object
        """

        return group[key][()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mmodel']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.16', 'h5py>=3.6.0', 'networkx>=2.8.3']

extras_require = \
{'docs': ['sphinx>=4.5.0,<5.0.0',
          'sphinx-book-theme>=0.3.2,<0.4.0',
          'nbsphinx>=0.8.8,<0.9.0'],
 'test': ['pytest>=7.1.1', 'pytest-cov>=3.0.0']}

setup_kwargs = {
    'name': 'mmodel',
    'version': '0.3.1',
    'description': 'Modular modeling framework for nonlinear scientific models',
    'long_description': 'MModel\n======\n\n|GitHub version| |PyPI version shields.io| |PyPI pyversions| |CircleCI|\n|Docs|\n\nMModel is a lightweight and modular model building framework\nfor small-scale and nonlinear models. The package aims to solve\nscientific program prototyping and distribution difficulties, making\nit easier to create modular, fast, and user-friendly packages.\nThe package is fully tested.\n\nQuickstart\n----------\n\nTo create a nonlinear model that has the result of\n`(x + y)log(x + y, base)`:\n\n.. code-block:: python\n\n    from mmodel import ModelGraph, Model, MemHandler\n    import math\n\n    def func_a(x, y):\n        return x + y\n\n    def func_b(sum_xy, base):\n        return math.log(sum_xy, base)\n\n    def func_c(sum_xy, log_xy):\n        return sum_xy * log_xy\n\n    # create graph links\n\n    grouped_edges = [\n        ("func a", ["func b", "func c"]),\n        ("func b", "func c"),\n    ]\n\n    node_objects = [\n        ("func a", func_a, ["sum_xy"]),\n        ("func b", func_b, ["log_xy"]),\n        ("func c", func_c, ["result"]),\n    ]\n\n    graph = ModelGraph(name="Example")\n    graph.add_grouped_edges_from(grouped_edges)\n    graph.set_node_objects_from(node_objects)\n\n    example_func = Model(graph, handler=MemHandler)\n\n    >>> print(example_func)\n    Example model\n      signature: base, x, y\n      returns: result\n      handler: MemHandler\n      modifiers: none\n\n    >>> example_func(2, 5, 3) # (5 + 3)log(5 + 3, 2)\n    24.0\n\nThe resulting ``example_func`` is callable.\n\nOne key feature of ``mmodel`` is modifiers, which modify callables post\ndefinition. To loop the "base" parameter.\n\n.. code-block:: python \n\n    from mmodel import subgraph_by_parameters, modify_subgraph, loop_modifier\n\n    subgraph = subgraph_by_parameters(graph, ["base"])\n    loop_node = Model(subgraph, MemHandler, [loop_modifier("base")])\n    looped_graph = modify_subgraph(graph, subgraph, "loop node", loop_node)\n    looped_model = Model(looped_graph, handler=MemHandler)\n\n    >>> print(looped_model)\n    Example model\n      signature: base, x, y\n      returns: result\n      handler: MemHandler\n      modifiers: none\n    \n    >>> looped_model([2, 4], 5, 3) # (5 + 3)log(5 + 3, 2)\n    [24.0, 12.0]\n\n\nModifiers can also be added to the whole model or a single node.\n\nTo draw the graph or the underlying graph of the model:\n\n.. code-block:: python\n    \n    graph.draw()\n    example_func.draw()\n\nInstallation\n------------\n\nGraphviz installation\n^^^^^^^^^^^^^^^^^^^^^\n\nTo view the graph, Graphviz needs to be installed:\n`Graphviz Installation <https://graphviz.org/download/>`_\nFor windows installation, please choose "add Graphviz to the\nsystem PATH for all users/current users" during the setup.\n\nMModel installation\n^^^^^^^^^^^^^^^^^^^^^^^\n\n.. code-block::\n\n    pip install mmodel\n\nDevelopment installation\n^^^^^^^^^^^^^^^^^^^^^^^^\nMModel uses `poetry <https://python-poetry.org/docs/>`_ as\nthe build system. The package works with both pip and poetry\ninstallation. \n\nTo install test and docs, despondencies run::\n\n    pip install .[test] .[docs]\n\nTo run the tests in different python environments and cases \n(py38, py39, coverage and docs)::\n\n    tox\n\nTo create the documentation, run under the "/docs" directory::\n\n    make html\n\n\n.. |GitHub version| image:: https://badge.fury.io/gh/peterhs73%2FMModel.svg\n   :target: https://github.com/peterhs73/MModel\n\n.. |PyPI version shields.io| image:: https://img.shields.io/pypi/v/mmodel.svg\n   :target: https://pypi.python.org/pypi/mmodel/\n\n.. |PyPI pyversions| image:: https://img.shields.io/pypi/pyversions/mmodel.svg\n\n.. |CircleCI| image:: https://circleci.com/gh/peterhs73/MModel.svg?style=shield\n    :target: https://circleci.com/gh/peterhs73/MModel\n\n.. |Docs| image:: https://img.shields.io/badge/Documentation--brightgreen.svg\n    :target: https://peterhs73.github.io/mmodel-docs/\n',
    'author': 'Peter Sun',
    'author_email': 'hs859@cornell.edu',
    'maintainer': 'Peter Sun',
    'maintainer_email': 'hs859@cornell.edu',
    'url': 'https://peterhs73.github.io/mmodel-docs/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import inspect
from mmodel.utility import parse_input, is_node_attr_defined, is_edge_attr_defined
from mmodel.draw import draw_graph
import networkx as nx


class Model:
    """Create model executable

    :param object graph: ModelGraph instance (digraph)
    :param class handler: Handler class that handles model execution. By default,
        the handler takes the graph as the first parameter. If additional parameters
        are required, use keyword arguments directly.
    :param list modifiers: modifiers used for the whole graph model executable.
        Optional, defaults to an empty list.
    """

    def __init__(self, graph, handler, modifiers: list = [], **handler_args):

        assert self._is_valid_graph(graph)

        self.__name__ = f"{graph.name} model"

        # store only the copy of the graph, note this is not the same copy
        # used by the handler
        # modify self._graph does not change the model itself
        self._graph = nx.freeze(graph.deepcopy())
        self._modifiers = modifiers
        self._handler = handler

        executor = handler(self._graph, **handler_args)
        self._handler_info = getattr(executor, "info", self._handler.__name__)

        for mdf in modifiers:
            executor = mdf(executor)

        self.__signature__ = executor.__signature__
        self.returns = executor.returns
        self.executor = executor

    def __call__(self, *args, **kwargs):

        # process input
        data_input = parse_input(self.__signature__, *args, **kwargs)

        return self.executor(**data_input)

    def __str__(self):
        """Output callable information"""

        sig_list = [str(param) for param in self.__signature__.parameters.values()]

        mod_str_list = [getattr(mod, "info", mod.__name__) for mod in self._modifiers]
        if mod_str_list:
            mod_str = ", ".join(mod_str_list)
        else:
            mod_str = "none"

        return "\n".join(
            [
                f"{self.__name__}",
                f"  signature: {', '.join(sig_list)}",
                f"  returns: {', '.join(self.returns)}",
                f"  handler: {self._handler_info}",
                f"  modifiers: {mod_str}",
                f"{self._graph.graph.get('doc', '')}",
            ]
        ).rstrip()

    @staticmethod
    def _is_valid_graph(G):
        """Check if model graph is valid to build an executable

        ``mmodel`` does not allow cycle graphs, graphs with isolated nodes,
        and all nodes have callable attributes defined.
        The method is bound to Model class because the features
        are specific to ``Model`` class.
        """

        assert nx.is_directed(G), "invalid graph: undirected graph"
        assert not nx.recursive_simple_cycles(G), "invalid graph: graph contains cycles"
        assert not list(nx.isolates(G)), "invalid graph: graph contains isolated nodes"

        assert is_node_attr_defined(
            G, "func"
        ), "invalid graph: graph contains nodes with undefined callables"

        # the following might occur when the node object is incorrectly constructed
        assert is_node_attr_defined(G, "returns"), (
            "invalid graph: graph contains nodes with undefined callables returns, "
            "recommend using ModelGraph set_node_object method to add node object"
        )
        assert is_node_attr_defined(G, "sig"), (
            "invalid graph: graph contains nodes with undefined callables signatures, "
            "recommend using ModelGraph set_node_object method to add node object"
        )
        assert is_edge_attr_defined(G, "val"), (
            "invalid graph: graph contains edges with undefined variable attributes, "
            "recommend using ModelGraph set_node_object method to add node object"
        )

        return True

    def get_node(self, node):
        """Quick access to node within the model"""

        return self._graph.nodes[node]

    def get_node_object(self, node):
        """Quick access to node callable within the model"""

        return self._graph.nodes[node]["func"]

    def view_node(self, node):
        """View a specific node"""

        return self._graph.view_node(node)

    def draw(self, method: callable = draw_graph):
        """Draw the graph of the model

        A drawing is provided. Defaults to ``draw_graph``
        """

        return method(self._graph, str(self))
